/****************************************************************************
 *
 * Portable Wireless Keyboard(TM) add-on driver for Xwindow v1.0
 * Copyright GPL Licence. 
 * 2005.6.04 lampo1 <lampo1@paran.com>
 *
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>
#include <fcntl.h>
#include <signal.h>
#include <termios.h>
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/extensions/XTest.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/time.h>

#define MAXNAME 16 
#define SHIFT 50    /* most linux systems assign keycode 50 to Shift_L */

/* One IR key can have multiple meaning by combinations */
/* with other modifier keys. */

struct keymapping {
    unsigned char irbyte;

    char nml[MAXNAME];
    unsigned int kcode_nml;

    char lfn[MAXNAME];
    unsigned int kcode_lfn;

    char rfn[MAXNAME];
    unsigned int kcode_rfn[2];
};  

struct keymapping KeyTable[51] = {
    {0x1b,        "Tab",0,             "",0,             "",{0,0}},
    {0x09,          "q",0,            "1",0,       "exclam",{0,0}},
    {0x0a,          "w",0,            "2",0,           "at",{0,0}},
    {0x0b,          "e",0,            "3",0,   "numbersign",{0,0}},
    {0x0c,          "r",0,            "4",0,       "dollar",{0,0}},
    {0x0d,          "t",0,            "5",0,      "percent",{0,0}},
    {0x0e,          "y",0,            "6",0,  "asciicircum",{0,0}},
    {0x3c,          "u",0,            "7",0,    "ampersand",{0,0}},
    {0x3d,          "i",0,            "8",0,     "asterisk",{0,0}},
    {0x3e,          "o",0,            "9",0,    "parenleft",{0,0}},
    {0x3f,          "p",0,            "0",0,   "parenright",{0,0}},
    {0x38,      "minus",0,  "bracketleft",0,    "braceleft",{0,0}},
    {0x32,  "BackSpace",0,             "",0,             "",{0,0}},
    {0x18,  "Caps_Lock",0,             "",0,             "",{0,0}},
    {0x11,          "a",0,             "",0,             "",{0,0}},
    {0x12,          "s",0,             "",0,             "",{0,0}},
    {0x13,          "d",0,             "",0,             "",{0,0}},
    {0x14,          "f",0,             "",0,             "",{0,0}},
    {0x15,          "g",0,             "",0,             "",{0,0}},
    {0x16,          "h",0,             "",0,             "",{0,0}},
    {0x44,          "j",0,             "",0,             "",{0,0}},
    {0x45,          "k",0,             "",0,             "",{0,0}},
    {0x46,          "l",0,             "",0,             "",{0,0}},
    {0x47,  "semicolon",0,             "",0,             "",{0,0}},
    {0x56, "apostrophe",0,        "grave",0,   "asciitilde",{0,0}},
    {0x57,     "Return",0,             "",0,             "",{0,0}},
    {0x58,    "Shift_L",0,             "",0,             "",{0,0}},
    {0x03,          "z",0,             "",0,             "",{0,0}},
    {0x10,          "x",0,             "",0,             "",{0,0}},
    {0x2c,          "c",0,             "",0,             "",{0,0}},
    {0x2d,          "v",0,             "",0,             "",{0,0}},
    {0x2e,          "b",0,             "",0,             "",{0,0}},
    {0x2f,          "n",0,             "",0,             "",{0,0}},
    {0x4c,          "m",0,             "",0,             "",{0,0}},
    {0x4d,      "comma",0,             "",0,             "",{0,0}},
    {0x4e,     "period",0,             "",0,             "",{0,0}},
    {0x48,         "Up",0,             "",0,        "Prior",{0,0}},
    {0x59,    "Shift_R",0,             "",0,             "",{0,0}},
    {0x1a,  "Control_L",0,             "",0,             "",{0,0}},
    {0x23,      "Alt_L",0,             "",0,             "",{0,0}},
    {0x17,      "space",0,             "",0,             "",{0,0}},
    {0x37,      "space",0,             "",0,             "",{0,0}},
    {0x51,       "Left",0,             "",0,         "Home",{0,0}},
    {0x52,       "Down",0,             "",0,         "Next",{0,0}},
    {0x53,      "Right",0,             "",0,          "End",{0,0}},
    {0x50,     "Delete",0,             "",0,             "",{0,0}},  
    {0x3a,      "slash",0,    "backslash",0,          "bar",{0,0}},
    {0x39,      "equal",0, "bracketright",0,   "braceright",{0,0}}, 
    {0x08,     "Escape",0,             "",0,             "",{0,0}}, /* Cmd, 49th */ 
    {0x22,           "",0,             "",0,             "",{0,0}}, /* Fn_L */
    {0x24,           "",0,             "",0,             "",{0,0}}  /* Fn_R */
};

int fd;
unsigned int auto_off_time = 300;    /* seconds */
unsigned int typematic_rate = 1;     /* n times of unit cycle */
unsigned int typematic_delay = 3;    /* n times of unit cycle */
unsigned int num_rptcode = 0;
unsigned char prev_key = 0;
unsigned int lfn_FLAG = 0;
unsigned int rfn_FLAG = 0;
Display *d;

void alarm_inthandler(int fd) {
    close(fd);
    exit(EXIT_SUCCESS);
}

void read_keymap() {
    FILE *fp;
    int i;
    int len = 254;
    char buf[255];
    char cset[] = " \n";
    char *p0, *p1, *p2;
 
    fp = popen("xmodmap -pke", "r");
    if ( fp == NULL ) {
        fprintf(stderr,"Cannot execute xmodmap -pke\n");
        exit(EXIT_FAILURE);
    }

    while ( fgets(buf, len, fp) ) {

	/* parsing one line of the result of xmodmap -pke */
	strtok(buf, cset);          /* discard "keycode" */
	p0 = strtok(NULL, cset);    /* point to keycode */
	strtok(NULL, cset);         /* discard "="  */
	p1 = strtok(NULL, cset);    /* point to X keysym name */ 
	p2 = strtok(NULL, cset);    /* point to the shifted X keysym name */

	for ( i = 0 ; i <= 48; i++) {

	    if ( (p1 != NULL) && (!strcmp( KeyTable[i].nml, p1)) ) {
		KeyTable[i].kcode_nml = atoi(p0);
	    }
	    
	    if ( (p1 != NULL) && (!strcmp( KeyTable[i].lfn, p1)) ) {
		KeyTable[i].kcode_lfn = atoi(p0);
	    }

	    if ( (p1 != NULL) && (!strcmp( KeyTable[i].rfn, p1)) ) {
		KeyTable[i].kcode_rfn[0] = atoi(p0);
		KeyTable[i].kcode_rfn[1] = 0;
	    }
	    
	    if ( (p2 != NULL) && (!strcmp( KeyTable[i].rfn, p2)) ) {
		KeyTable[i].kcode_rfn[0] = SHIFT;
		KeyTable[i].kcode_rfn[1] = atoi(p0);
	    }
	}
    } 
    pclose(fp);

    #ifdef DEBUG
    printf("In printing...\n");
    for ( i = 0; i <= 50; i++ ) {
	printf("#%d %x %s %d %s %d %s %d %d\n",
	       i,
	       KeyTable[i].irbyte,
	       KeyTable[i].nml,
	       KeyTable[i].kcode_nml,
	       KeyTable[i].lfn,
	       KeyTable[i].kcode_lfn,
	       KeyTable[i].rfn,
	       KeyTable[i].kcode_rfn[0],
	       KeyTable[i].kcode_rfn[1]
	       );
    }
    #endif	       
}
    
void read_rcfile(FILE *fp) {
  
    char dmy[255];
    while ( (fgets(dmy, sizeof(dmy), fp)) != NULL ) {

	if ( dmy[0] == '#' ) continue;
	else if ( dmy[0] == 'd' ) {
	    sscanf(dmy, "%*s %d", &typematic_delay);
	} else if ( dmy[0] == 'r' ) {
	    sscanf(dmy, "%*s %d", &typematic_rate);
	} else if ( dmy[0] == 'a' ) {
	    sscanf(dmy, "%*s %d", &auto_off_time);
	}
    }
    fclose(fp);
}

void translate( unsigned char ch) {

    int  i, j;
    unsigned int *ptr;

    unsigned char msb;   /* 0 : key pressed */
                         /* 1 means key released. */

    msb =  (ch & 128) >> 7 ;   
    ch &= 127 ; 

    for (i = 0; i <= 50 ; i++) {

	if ( ch == KeyTable[i].irbyte ) {
	    if ( ch == 0x22 ) {         /* Left-Fn */
		lfn_FLAG = (1^msb);
		break;
	    } else if ( ch == 0x24 ) {  /* Right-Fn */
		rfn_FLAG = (1^msb);
		break;
	    } else if ( lfn_FLAG == 1 ) {
		if ( ch == 0x50 ) {          /* LeftFn-Del */
		    close(fd);
		    exit(EXIT_SUCCESS);
		}
		if ( (ch == 0x46) || (ch == 0xc6) ) {       /* keymap reload */          
		    read_keymap();                               /* LeftFn-L */
		    return;
		}

		XTestFakeKeyEvent(d, KeyTable[i].kcode_lfn, (1^msb), 0);
		XFlush(d);
		break;
	    } else if ( rfn_FLAG == 1 ) {

		ptr = &KeyTable[i].kcode_rfn[0];

		if ( msb == 0 ) {       /* send kcodes in proper order */
		    for ( j = 0; j <= 1; j++) {
			if ( *ptr != 0 ) {
			    XTestFakeKeyEvent(d, *ptr, (1^msb), 0);
			    XFlush(d);

                            #ifdef DEBUG
			    printf("code= %d\n", *ptr);
                            #endif
			} 
			ptr++;
		    }
		    break;
		} else {                /* send kcodes in reverse order */
		    ptr++;
		    for ( j = 0; j <= 1; j++) {
			if ( *ptr != 0 ) {
			    XTestFakeKeyEvent(d, *ptr, (1^msb), 0);
			    XFlush(d);

                            #ifdef DEBUG
			    printf("code= %d\n", *ptr);
                            #endif
			} 
			ptr--;
		    }
		    break;
		} 		    
	    } else {
		XTestFakeKeyEvent(d, KeyTable[i].kcode_nml, (1^msb), 0);
		XFlush(d);
		break;
	    }
	}
    }
}

int process_irbyte( unsigned char ch) {
	
    if ( ch == 0x5c ) {                           /* key repetition code */
	if (  ((prev_key & 128) >> 7) == 0  ) {   /* if prev is make */
	    num_rptcode++;
	    if ( num_rptcode >= typematic_delay ) {
		translate(prev_key);
		num_rptcode--;
	    } 
	} /* if prev is break key, ignore. (any break key cannot repeat) */
	return 1;          
    }
       
    if ( ch == prev_key ) {
	return 1;          /* ignore one of the two duplicated break code */
    }

    num_rptcode = 0;
    translate( ch );
    prev_key = ch;
    return 1;
}
	
int main(int argc, char *argv[]) {

    unsigned char ch, irbyte;
    int nread, option;
    FILE *fp;
    char *device = "/dev/ttyS1";
    char *rcfile = "/etc/pwkrc";

    struct termios oldtio, newtio;
    struct sigaction newhandler;
    sigset_t blocked;

    while ( (option = getopt(argc, argv, "hd:c:")) != -1 ) {
	switch(option) {
	case 'h' :
	    fprintf(stderr,"Palm Wireless Keyboard driver v1.0\n");
	    fprintf(stderr,"Usage : pwkxwin [-d device] [-c rcfile]\n");
	    fprintf(stderr,"example : pwkxwin -d /dev/ttyS2 -c /etc/pwkrc\n");
	    exit(0);
	case 'd' :
	    device = optarg;
	    break;
	case 'c' :
	    rcfile = optarg;
	    break;
	default :
	    fprintf(stderr,"unknown option.\n");
	    exit(1);
	}
    }
	    
    /* signal install */
    newhandler.sa_handler = alarm_inthandler;
    sigemptyset(&blocked);
    sigaddset(&blocked, SIGALRM);
    newhandler.sa_mask = blocked;
    if ( sigaction(SIGALRM, &newhandler, NULL) == -1 ) {
	perror("sigaction install error\n");
	exit(EXIT_FAILURE);
    }

    /* open irda port device */
    fprintf(stderr, "Opening %s\n", device);
    fd = open(device, O_RDONLY | O_NOCTTY );
    if (fd == -1)  {
    	fprintf(stderr, "Cannot open %s\n", device);
    	exit(EXIT_FAILURE);
    }

    /* open rcfile */
    fprintf(stderr, "Opening %s\n", rcfile);
    fp = fopen( rcfile, "r");
    if ( fp == NULL ) {
	fprintf(stderr,"Cannot open %s\n", rcfile);
	fprintf(stderr,"Default time setting is applied.\n");
    } else {
	read_rcfile( fp );
    }

    /* read current xmodmap */
    read_keymap();

    /* Serial port setting to noncanonical, no echo, ...*/
    tcgetattr(fd, &oldtio); 
    bzero(&newtio, sizeof(newtio));
    newtio.c_cflag = B9600 | CRTSCTS | CS8 | CLOCAL | CREAD;
    newtio.c_iflag = IGNPAR;
    newtio.c_oflag = 0;
    newtio.c_lflag = 0;
    newtio.c_cc[VTIME]    = 0;   /* inter-character timer unused */
    newtio.c_cc[VMIN]     = 1;   /* blocking read until 1 chars received */
    tcflush(fd, TCIFLUSH);
    tcsetattr(fd,TCSANOW,&newtio);
               
    d = XOpenDisplay(NULL);
    
    while (1) {
	alarm( auto_off_time );

	nread = read(fd, &ch, 1);
        if ( ( nread > 0 ) && ( ch == 0xff ) ) {
      
	    nread = read(fd, &ch, 1);
	    if ( (nread > 0) && ( ch == 0xc0 ) ) {
	    
		nread = read(fd, &irbyte, 1);
		read(fd, &ch, 1);
		read(fd, &ch, 1); 
		nread = read(fd, &ch, 1);

		if ( ch == 0xc1 ) process_irbyte(irbyte);		
	    }
	} else if ( nread == -1 ) {
	    fprintf(stderr,"IrDA Port reading error.\n");
	    fprintf(stderr,"Recommend removing other irda modules.\n");
	    exit(EXIT_FAILURE);
	}         
    }  
}
